'''
Copyright (C) 2015 Rene Grzeszick. All rights reserved.
This file made available under the terms of the BSD license (see the Copyright.txt).

Created on Jun 30, 2014
'''
import numpy as np
import datetime, sys
from argparse import ArgumentParser

from sklearn.svm import SVC
from sklearn import naive_bayes
from sklearn.ensemble import RandomForestClassifier

from config import SVWConfig
from dataset_utils import DatasetOrganizer
from SpatialVisualWords import SpatialVisualWords

from file_handling.os_wrapper import OsUtils
from file_handling.serialization_io import SerializationIO


#--------------------
# Train routines
#--------------------

def compute_all_features(config):
    '''
    Computes the training features
    @param config: SVWConfig file
    '''
    feature_type = config.descriptor_type    
    
    # Set paths
    data_path = config.data_path()
    feature_path = data_path + "/" + config.feature_filename()
    meta_info_path = config.feature_info_path()
    
    # Create feature extractor
    feature_extractor = SpatialVisualWords(meta_info_path, config.get_spatial_feature_type(),
                                           config.sift_step, config. sift_sizes, 
                                           config.spatial_partitions, config.vocab_size)
    
    # Compute or load a codebook
    codebook_path = data_path + "/" + config.codebook_filename()
    if not OsUtils.is_file(data_path, config.codebook_filename()):
        print("Training BoF codelibrary")
        feature_extractor.cluster(config.train_path())
        feature_extractor.store_codebook(codebook_path)
    else:
        feature_extractor.load_codebook(codebook_path)
    
    # Iterate all classes and files
    feature_dim = feature_extractor.get_dimension()
    class_list = OsUtils.get_dirlisting(data_path, dirs_only=True)
    for c_i, class_name in enumerate(class_list):
        
        # Skip categories where the features have already been computed
        if OsUtils.is_file(feature_path + str(c_i)):
            continue
    
        # Create feature matrix for category
        file_list = OsUtils.get_dirlisting(data_path + "/" + class_name)
        count = len(file_list)
        features = np.zeros((count, feature_dim))
        
        print("Computing features for class " + str(class_name))
        # Compute features for all files
        file_list = OsUtils.get_dirlisting(data_path + "/" + class_name)
        for f_i, file_name in enumerate(file_list):
            file_path = data_path + "/" + class_name + "/" + file_name
            features[f_i, :] = feature_extractor.get_features(file_path)
            f_i += 1
        
        # Serialize features for this category
        SerializationIO.store_data(features, feature_path + str(c_i))
        
    # Concatenate data
    f_i = 0
    features = np.zeros((config.all_image_count(), feature_dim))
    for c_i, class_name in enumerate(class_list):  
        cat_features = SerializationIO.load_data(feature_path + str(c_i))
        l_c = cat_features.shape[0]
        features[f_i:f_i+l_c] = cat_features
        f_i += l_c
        
    # Store as one matrix
    SerializationIO.store_data(features, feature_path)
    
    # Finally delete buffered data
    for c_i, _ in enumerate(class_list):  
        OsUtils.delete_file(feature_path + str(c_i))
    return features


def split_train_features(config):
    '''
    Reads the pre-computed features
    Then splits them into a train and test set.
    @param config: SVWConfig file
    '''
    
    # Create paths
    feature_name = config.feature_filename()
    data_path = config.data_path()
    train_path = config.train_path()
    test_path = config.test_path()
    
    feature_path = data_path + "/" + feature_name
    train_feature_path = train_path + "/" + feature_name
    test_feature_path = test_path + "/" + feature_name
    
    all_features = SerializationIO.load_data(feature_path)
    train_idx = np.zeros(all_features.shape[0], dtype=bool)
    
    i = 0
    class_list = OsUtils.get_dirlisting(data_path)
    for class_name in class_list:
        # Skip files
        if not OsUtils.is_dir(data_path, class_name):
            continue
        # Compute features for all files
        file_list = OsUtils.get_dirlisting(data_path + "/" + class_name)
        for f_name in file_list:
            if OsUtils.is_file(train_path + "/" + class_name, f_name):
                train_idx[i] = True
            i += 1
    
    # Serialize train features
    SerializationIO.store_data(all_features[train_idx], train_feature_path)
    # Invert train idx in order to get test indizes
    train_idx = (train_idx == False)
    SerializationIO.store_data(all_features[train_idx], test_feature_path)
    return

def get_train_labels(config):
    '''
    Returns a label list
    @param config: SVWConfig file
    '''
    
    # Set train path
    train_path = config.train_path()
    
    # Iterate all classes and files
    labels = []
    class_list = OsUtils.get_dirlisting(train_path, dirs_only=True)
    c_i = 0
    for class_name in class_list:
        #Append files
        file_list = OsUtils.get_dirlisting(train_path + "/" + class_name)
        for _ in range(len(file_list)):
            labels.append(c_i)
        c_i += 1
        
    # Create label matrix
    labels = np.array(labels)
    return labels

def train_sklearn_classifier(config, train_features):
    '''
    Trains a sklean classifier (e.g. bayes/forest), stores the model and returns the classifier
    @param config: SVWConfig file
    @param train_features: Numpy matrix containing the features for each train file 
    '''
    
    # Get training labels
    train_labels = get_train_labels(config)
    
    # Init classifier
    if config.classifier == "RandomForest":
        classifier = RandomForestClassifier(n_estimators=1000, criterion="entropy", 
                                                 max_depth=15, n_jobs=1)
    elif config.classifier == "Bayes":
        classifier = naive_bayes.MultinomialNB(alpha=0.05)
    elif config.classifier == "SVM":
        classifier = SVC(probability=True, kernel='linear') 
    
    # Train classifier
    classifier.fit(train_features, train_labels)
    return classifier


#--------------------
# Test routines
#--------------------

def print_testresults(correct, false, class_list):
    '''
    Prints the evaluation results
    @param correct: List of correct classifications per class
    @param false: List of false classifications per class
    @param classlist: List of classnames 
    '''
    print("Correct predictions:")
    print(sum(correct))
    print("False predictions:")
    print(sum(false))
    print("Recognition rate:")
    rec_rate = float(sum(correct))/(sum(correct)+sum(false))
    rec_rate = round(rec_rate, 3)
    print(rec_rate)
    
    '''
    # Additional print outs for class specific rates
    avg_over_class = 0
    print("Class specific rates:")
    for i in range(len(correct)):
        rec_rate = float(correct[i])/(correct[i]+false[i])
        rec_rate = round(rec_rate, 3)
        print(class_list[i] + ": " + str(rec_rate))
        avg_over_class += rec_rate
    
    avg_over_class /= len(correct)
    print("Class average:", avg_over_class)
    '''
    
    return

def eval_test(config, classifier):
    '''
    Evaluates on the test set
    @param config: SVWConfig file
    @param classifier: Trained sklearn classifier 
    '''
    
    # Set paths
    test_path = config.test_path()
    feature_name = config.feature_filename()
    
    # Create feature extractor and/or read feature matrix
    if not OsUtils.is_file(test_path, feature_name):
        raise RuntimeError("No test features found")
    
    test_features = SerializationIO.load_data(test_path + "/" + feature_name)
    
    # Iterate all classes and files
    class_list = OsUtils.get_dirlisting(test_path, dirs_only=True)
    c_i = 0
    i = 0
    correct = np.zeros((config.category_count()))
    false = np.zeros((config.category_count()))
    
    for class_name in class_list:
        print("Evaluating class " + str(class_name))
        # Evaluate files
        file_list = OsUtils.get_dirlisting(test_path + "/" + class_name)
        for file_name in file_list:
            # Predict label
            features = test_features[i]
            label = classifier.predict(features)
            i += 1
            
            # Count true/false            
            if label == c_i:
                correct[c_i] += 1
            else:
                false[c_i] += 1
        # Increase class count       
        c_i += 1
    
    # Print out the results
    print_testresults(correct, false, class_list)
    
    return



#--------------------
# Run
#--------------------

def run(config):
    '''
    Runs the experiment
    '''

    #
    # Splits the data into test and train
    #
    train_path = config.train_path()
    if not OsUtils.is_dir(train_path):
        # Split images
        trainfile_list = DatasetOrganizer.split_rand_test_train(config.data_path(), config.number_train_images)
        # Create a file containing the training list
        train_info_path = config.root_path
        train_info_filename = "train" + str(config.number_train_images)
        DatasetOrganizer.save_train_files(train_info_path, train_info_filename, trainfile_list)
        
        
    #
    # Compute all features
    #
    feature_file = config.feature_filename()
    if not OsUtils.is_file(config.data_path(), feature_file):
        compute_all_features(config)
        split_train_features(config)
    
    # Load training features
    feature_path = train_path + "/" + feature_file
    train_features = SerializationIO.load_data(feature_path)
    
    #
    # Train or load classifier
    #  
    clfc_path = train_path + "/" + config.classifier_filename()
    if not OsUtils.is_file(clfc_path):
        classifier = train_sklearn_classifier(config, train_features)
        SerializationIO.store_data(classifier, clfc_path)
    else:
        classifier = SerializationIO.load_data(clfc_path) 
        
        
    eval_test(config, classifier)
    
    return

def parse_args(argv):
    '''Parses input arguments'''
    print("Command line arguments are:")
    print(argv)
    
    
    parser = ArgumentParser()
    parser.add_argument('--root_path', required=True, 
                        help='Dataset root dir')
    parser.add_argument('--descriptor_type', required=False, default="SIFTxy",
                        help='Type of feature SIFT|SIFTxy|SIFTradial')
    parser.add_argument('--vocab_size', type=int, required=False, default=1000,
                        help='Size of the visual vocabulary')
    parser.add_argument('--sift_step', type=int, required=False, default=8,
                        help='Stepsize of the SIFT dense grid')
    parser.add_argument('--sift_sizes', nargs='+', type=int, required=False, default=4,
                        help='SIFT descriptor bin size(s), note that the SIFT descriptor has 4x4 bins')
    parser.add_argument('--spatial_partitions',  nargs='+', type=int, required=False, default=(2)
                        , help='Optional: Number of spatial partitions')
    parser.add_argument('--rooted_histograms',  type=bool, required=False, default=False, 
                        help='Optional: Use square rooted histograms')
    parser.add_argument('--classifier', required=False, default="SVM", 
                        help='Optional: sklearn classifier SVM|Bayes|RandomForest')
    
    parsed_args = parser.parse_args(argv)
    parsed_args = vars(parsed_args)
    config = SVWConfig(**parsed_args)
    
    return config

if __name__ == '__main__':
    '''
    Runs an experiment on an image database:
    The expected format is root_path/data/[categories]
    
    The following steps will be performed:
    * The dataset is randomly split into a test and train dir
    * The features for all images are computed using a dense grid of SIFT descriptors 
      and a Bag-of-Features based on Spatial Visual Words
    * The features are split into train and test accordingly
    * A classifier is trained on the training set
    * The classifier is evaluated on the test set
    '''
    config = parse_args(sys.argv[1:])
    run(config)   
    
