'''
Copyright (C) 2015 Rene Grzeszick. All rights reserved.
This file made available under the terms of the BSD license (see the Copyright.txt).

Created on Jun 30, 2014
'''

import random
import numpy as np

from file_handling.os_wrapper import OsUtils
from bof.sift_features import DenseSiftExtractor
from bof.cdist_vq import VectorQuantizer
from bof.cdist_lloyd import LloydClusterer
from bof.spatial_features import SpatialFeatureCreator

from file_handling.serialization_io import SerializationIO

class SpatialVisualWords(object):
    '''
    Computes a Bag-of-Features representation using Spatial Visual Words
    '''    
    
    def __init__(self,
                 meta_info_path,
                 spatial_type=None,
                 grid_step=5,
                 sift_sizes=(4,6,8,10),
                 spatial_partitions=2,
                 vocab_size=1000,
                 cluster_coreset_size=200000):
        '''
        @param meta_info_path: Path to store metadata that is used
        @param feature_type: Use SIFT|SURF features
        @param grid_step:    Step size of the used dense grid
        @param sift_sizes:   Sizes of the SIFT bins (sqrt(step) for surf)
        @param spatial_type: Use xy|rad spatial quantization
        @param spatial_partitions: Number of spatial partitions
        @param vocab_size: Size of the visual vocabulary
        @param cluster_coreset_size: Use this parameter to limit the number of features that are 
                                     considered for computing the visual vocabulary
        '''
        
        assert(spatial_type in ["xy", "radial", None])
        
        self.__sift_extractor = DenseSiftExtractor(sift_sizes, grid_step)
        self.__surf_extractor = DenseSiftExtractor(grid_step)
        self.__spatial_trafo = SpatialFeatureCreator(spatial_type, spatial_partitions, meta_info_path)
        self.__lloyd = LloydClusterer(vocab_size, max_iter=10)
        self.__vocab_size = vocab_size
        self.__codebook = None
        self.__coreset_size = cluster_coreset_size
        return
     
    def get_dimension(self):
        dimension =  self.__vocab_size
        return dimension
       
    def cluster(self, train_img_path):
        '''
        Clusters the training data in order to compute a visual vocabulary
        @param train_img_path: Path to train image dir 
        '''
        
        #
        # Iterate all training files
        # Create a file list
        #
        class_list = OsUtils.get_dirlisting(train_img_path, dirs_only=True)
        image_list = []
        
        for class_name in class_list:               
            class_images = OsUtils.get_dirlisting(train_img_path + "/" + class_name)
            for image_name in class_images:
                # Extract features
                img_path = train_img_path + "/" + class_name + "/" + image_name
                image_list.append(img_path)
        
        #Randomize image list
        random.shuffle(image_list)      
        
        feature_list = None
        desc_count = 0  #descriptor count
        # Process images
        for img_path in image_list:
            
            # Extract features
            localizations, descriptors = self.__sift_extractor.get_phow_features(img_path, use_smoothing=False)
            # Transpose data
            localizations = np.array(localizations).T
            descriptors = np.array(descriptors).T
            
            # Add spatial component
            descriptors = self.__spatial_trafo.add_spatial_feature(localizations, descriptors, randomize=True)
            descriptors = np.array(descriptors, 'f')
            
            #For performance reasons a maximum coreset size is defined
            if desc_count+descriptors.shape[0] >= self.__coreset_size:
                break
            
            # Store all features in one feature matrix
            if feature_list == None:
                feature_list = np.zeros((self.__coreset_size, descriptors.shape[1]))
            feature_list[desc_count:desc_count+descriptors.shape[0]] = descriptors 
            desc_count += descriptors.shape[0]
                
        if desc_count < self.__coreset_size:
            feature_list = feature_list[:desc_count]
            
        #                
        # Compute visual vocabulary
        #
        codebook, labels = self.__lloyd.cluster(feature_list)
        self.__codebook = codebook
        del labels
        del feature_list
        
        return
    
    def store_codebook(self, path):
        '''Stores the codebook at the given path'''
        if self.__codebook == None:
            raise RuntimeError("Codebook has not been set")
        
        SerializationIO.store_data(self.__codebook, path)
        return
        
    def load_codebook(self, path):
        '''Loads the codebook at the given path'''
        self.__codebook = SerializationIO.load_data(path)
        return
    
    def has_codebook(self):
        ''' Is the codebook set? '''
        return self.__codebook != None
    
    def __encode_bof(self, id_list):
        '''Computes a BoF hist using bincount'''
        term_vector = np.zeros((self.__vocab_size))
        term_vector_sub = np.bincount(id_list)
        term_vector[0:len(term_vector_sub)] = term_vector_sub
        del term_vector_sub
        return term_vector

    def __quantize(self, img_path, sqrooted=True):
        '''
        Returns a Bag of Features using Spatial Visual Words
        @param img: PIL input Image        
        '''
        
        # Extract features
        localizations, descriptors = self.__sift_extractor.get_phow_features(img_path, use_smoothing=False)
        # Transpose data
        localizations = np.array(localizations).T
        descriptors = np.array(descriptors).T
        # Add spatial component
        descriptors = self.__spatial_trafo.add_spatial_feature(localizations, descriptors)
       
        # Quantize features
        vector_quantizer = VectorQuantizer(self.__codebook)
        id_list = vector_quantizer.quantize(descriptors)
        term_vector = self.__encode_bof(id_list)
        
        if sqrooted:
            # Normalize term vector to frequencies
            term_vector /= sum(term_vector)
            # Compute square rooted histogram: Hellinger Kernel
            term_vector = np.sqrt(term_vector)                
        
        return term_vector
    
    def get_features(self, path):
        '''
        Returns a Bag of Features using Spatial Visual Words
        @param path: Image path
        @precondition: Codebook must be set
        '''
        if self.__codebook == None:
            raise RuntimeError("Codebook has not been set")
        
        return self.__quantize(path)
