'''
Copyright (C) 2015 Rene Grzeszick. All rights reserved.
This file made available under the terms of the BSD license (see the Copyright.txt).

Created on Jun 5, 2013
'''


import numpy as np
import scipy.spatial.distance as ssd

class LloydClusterer(object):
    '''
    Class object for clustering data using Lloyds k-Means implementation.
    This implementation uses cdist and can therefore use different distance metrics.
    It is necessary to fit all data into the memory.
    '''


    def __init__(self, vocab_size, max_iter=25, verbose=False):
        '''
        Initializes the generalized lloyd algorithm.
        @param vocab_size:    Clusters to compute by the kmeans algorithm
        @param max_iter:    Number of iterations
        '''
        
        if vocab_size <= 0:
            raise ValueError("Number of centroids to compute must be > 0")
        
        self.__vocab_size = vocab_size
        self.__iterations = max_iter
        self.__verbose = verbose
        
        return
    
    
    def __intersect2d(self, arr1, arr2):
    
        arr1_view = arr1.view([('', arr1.dtype)]*arr1.shape[-1])
        arr2_view = arr2.view([('', arr2.dtype)]*arr2.shape[-1])
        
        intersected = np.intersect1d(arr1_view, arr2_view)
        return intersected.view(arr1.dtype).reshape(-1, arr1.shape[-1])

    
    def __init_ids(self, data, vocab_size):
        '''
        Get a list of random indices for the lloyd initialization
        * Avoids same random indices
        * Or same feature vectors
        '''
        rand_ints = []
        len_data = len(data)
        
        while len(rand_ints) < vocab_size:
            i = np.random.randint(0, len_data)
            if not i in rand_ints \
               and len(self.__intersect2d(data[i], data[rand_ints])) == 0:
                rand_ints.append(i)
            
        return rand_ints
    
    
    def cluster(self, data):
        '''
        Runs the k-Means clustering
        @param data: Requires the data in a [Features x Dimensions] format.
        @return: codebook, labels for each observation
        '''
                
        vocab_size = self.__vocab_size
        iterations = self.__iterations
        verbose = self.__verbose
        
        assert(len(data) >= vocab_size)
        
        if verbose:
            print("Starting Lloyd clustering, this may take a while ...")
            print("Initializing")
        
        # Take random starting points from data array
        # implicitely same datatype
        rand_ids = self.__init_ids(data, vocab_size)
        codebook = np.array(data[rand_ids])
        len_d = len(data)

        for i in range(iterations):
            
            if verbose:
                print("Starting iteration: " + str(i) + " of " + str(iterations) + " ...")
                        
            labels = np.zeros(len_d)
            for d_i in range(len_d):
                dists = ssd.cdist([data[d_i]], codebook)
                labels[d_i] = np.argmin(dists)
                #delete distances since they allocate a lot of memory and are no longer needed
                del dists
            
            #Compute the new centroid for each partition
            empty_clusters = 0
            for p_i in range(vocab_size):
                partition_ids = (labels == p_i)
                partition_data = data[partition_ids, :]
                if len(partition_data) > 0:
                    codebook[p_i] = np.mean(data[partition_ids, :], axis=0)
                else:
                    #Error case: Keep centroid to avoid division by zero
                    empty_clusters += 1
                
            #Free data
            del partition_data
            del partition_ids
                        
            if verbose:
                print("...finished.")
                print(str(empty_clusters) + " clusters are empty.")
            
            
            
        return codebook, labels
