'''
Copyright (C) 2015 Rene Grzeszick. All rights reserved.
This file made available under the terms of the BSD license (see the Copyright.txt).

Created on Jun 5, 2013
'''


import numpy as np
import scipy.spatial.distance as ssd

class VectorQuantizer(object):
    '''
    A vector quantizer
    '''


    def __init__(self, codebook):
        '''
        Initialize the vector quantizer
        @param codebook:    Codebook that is used for the quantization
        '''
        
        if codebook is None:
            raise ValueError("codebook must not be None")
        
        self.__codebook = codebook
        self.__dfunction = "euclidean"
        
        return
    
    def quantize(self, data):
        '''
        Quantizates the data.
        @param data: List of observations in the format [features x dimensions]
        @return: lists of labels
        '''
        
        codebook = self.__codebook
        dfunction = self.__dfunction
        
        dists = ssd.cdist(data, codebook, dfunction)
        min_ids = np.argmin(dists, axis=1)
            
        return min_ids
    
    def set_distance(self, dfunction):
        '''
        Sets a distance parameter that is used for the quantization.
        @param distance: Name of a distance function or parameter 
        
        The following are common calling conventions:

        1. ``Y = cdist(XA, XB, 'euclidean')``
    
           Computes the distance between :math:`m` points using
           Euclidean distance (2-norm) as the distance metric between the
           points. The points are arranged as :math:`m`
           :math:`n`-dimensional row vectors in the matrix X.
    
        2. ``Y = cdist(XA, XB, 'minkowski', p)``
    
           Computes the distances using the Minkowski distance
           :math:`||u-v||_p` (:math:`p`-norm) where :math:`p \geq 1`.
    
        3. ``Y = cdist(XA, XB, 'cityblock')``
    
           Computes the city block or Manhattan distance between the
           points.
    
        4. ``Y = cdist(XA, XB, 'seuclidean', V=None)``
    
           Computes the standardized Euclidean distance. The standardized
           Euclidean distance between two n-vectors ``u`` and ``v`` is
    
           .. math::
    
              \sqrt{\sum {(u_i-v_i)^2 / V[x_i]}}.
    
           V is the variance vector; V[i] is the variance computed over all
              the i'th components of the points. If not passed, it is
              automatically computed.
    
        5. ``Y = cdist(XA, XB, 'sqeuclidean')``
    
           Computes the squared Euclidean distance :math:`||u-v||_2^2` between
           the vectors.
    
        6. ``Y = cdist(XA, XB, 'cosine')``
    
           Computes the cosine distance between vectors u and v,
    
           .. math::
    
              \frac{1 - uv^T}
                   {{|u|}_2 {|v|}_2}
    
           where :math:`|*|_2` is the 2-norm of its argument *.
    
        7. ``Y = cdist(XA, XB, 'correlation')``
    
           Computes the correlation distance between vectors u and v. This is
    
           .. math::
    
              \frac{1 - (u - n{|u|}_1){(v - n{|v|}_1)}^T}
                   {{|(u - n{|u|}_1)|}_2 {|(v - n{|v|}_1)|}^T}
    
           where :math:`|*|_1` is the Manhattan (or 1-norm) of its
           argument, and :math:`n` is the common dimensionality of the
           vectors.
    
        8. ``Y = cdist(XA, XB, 'hamming')``
    
           Computes the normalized Hamming distance, or the proportion of
           those vector elements between two n-vectors ``u`` and ``v``
           which disagree. To save memory, the matrix ``X`` can be of type
           boolean.
    
        9. ``Y = cdist(XA, XB, 'jaccard')``
    
           Computes the Jaccard distance between the points. Given two
           vectors, ``u`` and ``v``, the Jaccard distance is the
           proportion of those elements ``u[i]`` and ``v[i]`` that
           disagree where at least one of them is non-zero.
    
        10. ``Y = cdist(XA, XB, 'chebyshev')``
    
           Computes the Chebyshev distance between the points. The
           Chebyshev distance between two n-vectors ``u`` and ``v`` is the
           maximum norm-1 distance between their respective elements. More
           precisely, the distance is given by
    
           .. math::
    
              d(u,v) = \max_i {|u_i-v_i|}.
    
        11. ``Y = cdist(XA, XB, 'canberra')``
    
           Computes the Canberra distance between the points. The
           Canberra distance between two points ``u`` and ``v`` is
    
           .. math::
    
             d(u,v) = \sum_u \frac{|u_i-v_i|}
                                  {(|u_i|+|v_i|)}
    
    
        12. ``Y = cdist(XA, XB, 'braycurtis')``
    
           Computes the Bray-Curtis distance between the points. The
           Bray-Curtis distance between two points ``u`` and ``v`` is
    
    
           .. math::
    
                d(u,v) = \frac{\sum_i (u_i-v_i)}
                              {\sum_i (u_i+v_i)}
    
        13. ``Y = cdist(XA, XB, 'mahalanobis', VI=None)``
    
           Computes the Mahalanobis distance between the points. The
           Mahalanobis distance between two points ``u`` and ``v`` is
           :math:`(u-v)(1/V)(u-v)^T` where :math:`(1/V)` (the ``VI``
           variable) is the inverse covariance. If ``VI`` is not None,
           ``VI`` will be used as the inverse covariance matrix.
    
        14. ``Y = cdist(XA, XB, 'yule')``
    
           Computes the Yule distance between the boolean
           vectors. (see yule function documentation)
    
        15. ``Y = cdist(XA, XB, 'matching')``
    
           Computes the matching distance between the boolean
           vectors. (see matching function documentation)
    
        16. ``Y = cdist(XA, XB, 'dice')``
    
           Computes the Dice distance between the boolean vectors. (see
           dice function documentation)
    
        17. ``Y = cdist(XA, XB, 'kulsinski')``
    
           Computes the Kulsinski distance between the boolean
           vectors. (see kulsinski function documentation)
    
        18. ``Y = cdist(XA, XB, 'rogerstanimoto')``
    
           Computes the Rogers-Tanimoto distance between the boolean
           vectors. (see rogerstanimoto function documentation)
    
        19. ``Y = cdist(XA, XB, 'russellrao')``
    
           Computes the Russell-Rao distance between the boolean
           vectors. (see russellrao function documentation)
    
        20. ``Y = cdist(XA, XB, 'sokalmichener')``
    
           Computes the Sokal-Michener distance between the boolean
           vectors. (see sokalmichener function documentation)
    
        21. ``Y = cdist(XA, XB, 'sokalsneath')``
    
           Computes the Sokal-Sneath distance between the vectors. (see
           sokalsneath function documentation)
    
    
        22. ``Y = cdist(XA, XB, 'wminkowski')``
    
           Computes the weighted Minkowski distance between the
           vectors. (see sokalsneath function documentation)
    
        23. ``Y = cdist(XA, XB, f)``
    
           Computes the distance between all pairs of vectors in X
           using the user supplied 2-arity function f. For example,
           Euclidean distance between the vectors could be computed
           as follows::
    
             dm = cdist(XA, XB, (lambda u, v: np.sqrt(((u-v)*(u-v).T).sum())))
    
           Note that you should avoid passing a reference to one of
           the distance functions defined in this library. For example,::
    
             dm = cdist(XA, XB, sokalsneath)
    
           would calculate the pair-wise distances between the vectors in
           X using the Python function sokalsneath. This would result in
           sokalsneath being called :math:`{n \choose 2}` times, which
           is inefficient. Instead, the optimized C version is more
           efficient, and we call it using the following syntax.::
    
             dm = cdist(XA, XB, 'sokalsneath')'''
        
        self.__dfunction = dfunction
        return
