'''
Copyright (C) 2015 Rene Grzeszick. All rights reserved.
This file made available under the terms of the BSD license (see the Copyright.txt).

Created on Nov 9, 2012
'''

import Image
import numpy as np
import vlfeat

class BoundingBox():
    
    
    def __init__(self, row_min, col_min, row_max, col_max, annotation=None):
        self.__annotation = annotation
        self.__row_min = row_min
        self.__row_max = row_max
        self.__col_min = col_min
        self.__col_max = col_max
        
    def get_box_array(self):
        box_array = [self.__row_min, self.__col_min, self.__row_max, self.__col_max]
        return np.array(box_array, 'f')
    

class DenseSiftExtractor(object):
    '''
    vl_phow Dense SIFT extractor
    
    Extracts SIFT features with an upward orientation at different scales on a grid.
    Uses the vl_feat dense SIFT operation.    
    '''
    
    # Constant variables:
    __color = 'gray' 
    __numchannels = 1 

    def __init__(self, sizes=(4, 6, 8, 10), step=2, fast=True, windowsize=1.5, contrastthreshold=0.005, verbose=False):
        '''
        step    Extract a descriptor each STEP pixels.
        size    A spatial bin covers SIZE pixels.
        norm    Append the frames with the normalization factor applied to 
                   each descriptor. In this case, F has 3 rows and this value 
                   is the 3rd row. This information can be used to suppress
                   descriptors with low contrast.
        fast       Use a flat rather than Gaussian window. Much faster.
        windowsize    Size of the Gaussian window
        verbose    Be verbose. 
        '''
        if isinstance(sizes, int):
            sizes = [sizes]
        self.__sizes = sizes
        self.__step = step
        self.__fast = fast
        self.__windowsize = windowsize
        self.__contrastthreshold = contrastthreshold
        self.__verbose = verbose
        
        return
    
    
    def __vl_dsift(self, I, bounds, bin_size, norm=False):
        '''
        Extract dense SIFT features given
        @param I: The image 
        @param bounds: [row_min, col_min, row_max, col_max]
        @param bin-size: Size of the SIFT bin
        @param norm: Return the contrast value as part of f or not
        
        @return: [f, d] frames and descriptors
        '''
        [f, d] = vlfeat.vl_dsift(I, step=self.__step, bounds=bounds, size=int(bin_size), fast=self.__fast, windowsize=self.__windowsize, verbose=self.__verbose, norm=norm)
        return [f, d]
    
    @staticmethod
    def __normalize_image(image):
        '''
        Converts the image to gray scale float32, [0..1], 
        F continuous (column major order)
        ATTENTION if the maximal value > 1.0 assume original range [0..255]
                  if the maximal value <=1.0 assume original range [0..1] 
        @param image: a numpy ndarray (NxMx1 or NxMx3) 
        '''
        image = np.array(image, dtype='float32', order='F')  # 'F' = column-major order!
        
        if image.ndim == 3:
            image = 0.2989 * image[:, :, 0] + 0.5870 * image[:, :, 1] + 0.1140 * image[:, :, 2]
        
        if image.ndim != 2:
            raise ValueError('Unsupported number of dimensions in image!')
            
        image_vmax = image.max()
        if image_vmax > 1.0:
            print 'ATTENTION: Normalizing image from [0..255] to [0..1]'
            print '  --> estimated [0..255] due to maximal intensity value: %.2f' % image_vmax
            image = image / 255.0
            
        return image
    
            
    def get_phow_features(self, path, use_smoothing=True, magnif=6):
        '''
        Extracts the dense SIFT features using the phow scheme
        Please note that so far the computation uses gray values only
        
        @param use_smoothing: Use gaussian smoothing for computing a scale space
                              with respect to the different SIFT bin sizes
        @param magnif:        Magnification factor used for computing the variance 
                              at different scales: sigma * magnif = size
        @return: a tuple of locations and descriptors: l, d
        '''
                
        # 
        # Set gray value processing
        # 
        if self.__color == 'gray':
            I = Image.open(path)
            I = vlfeat.vl_rgb2gray(np.array(I))    
            I = np.array(I, 'f', order='F')  # 'F' = column-major order!
        
        locations = []
        extractedFeatures = []
        
        #
        # Compute features for all given sizes
        #
        for si in self.__sizes:

            #
            # Recall from VL_DSIFT() that the first descriptor for scale SIZE has
            # center located at XC = XMIN + 3/2 SIZE (the Y coordinate is
            # similar). It is convenient to align the descriptors at different
            # scales so that they have the same geometric centers. For the
            # maximum size we pick XMIN = 1 and we get centers starting from
            # XC = 1 + 3/2 MAX(OPTS.SIZES). For any other scale we pick XMIN so
            # that XMIN + 3/2 SIZE = 1 + 3/2 MAX(OPTS.SIZES).
            #
            # In practice, the offset must be integer ('bounds'), so the
            # alignment works properly only if all OPTS.SZES are even or odd.
            #
            
            off = np.floor(1 + 3.0 / 2.0 * (max(self.__sizes) - si)) 
            
        
            #
            # smooth the image to the appropriate scale based on the size
            # of the SIFT bins
            #
            if use_smoothing:
                sigma = float(si) / magnif 
                ims = vlfeat.vl_imsmooth(I, sigma)
            else:
                ims = I
            
            #
            # Create proper bounds for the SIFT features
            #
            bounding_box = BoundingBox(off, off, ims.shape[0] - off, ims.shape[1] - off)
            bounds = bounding_box.get_box_array()
            
            norm = (self.__contrastthreshold > 0)
            
            
            #
            # extract dense SIFT features: frame, descriptors
            # %%%from all channels
            # for k in range(self.__numchannels): %%%
            #
            [f, d] = self.__vl_dsift(ims, bounds, si, norm)
            
            #
            # Store x,y locations and extracted features in two matrices:
            #
            if norm:
                contrast = f[2, :]
                f = f[:2, :]
                d[:, contrast < self.__contrastthreshold] = 0 
            
            if extractedFeatures == []:
                locations = f
                extractedFeatures = d
            else:
                locations = np.concatenate((locations, f), axis=1)
                extractedFeatures = np.concatenate((extractedFeatures, d), axis=1)
        
    
        return locations, extractedFeatures
