'''
Copyright (C) 2015 Rene Grzeszick. All rights reserved.
This file made available under the terms of the BSD license (see the Copyright.txt).

Created on Jan 16, 2013
'''


import types
import os
import numpy as np
from scipy.spatial.distance import cdist

from file_handling.serialization_io import SerializationIO

class SpatialFeatureCreator(object):
    '''
    This class takes a set of appearance features a and appends spatial features s to them
    It returns v = (a_0, .. , a_n, s_0, .. , s_n). Used for creating Spatial Visual Vocabularies.
    '''
    
    #
    # Feature type
    #
    __valid_features = [None, "xy", "radial"]
    __feature_type = None

    __feature_info_path = None
    __feature_average_sum = None

    #
    # Quantization values
    #
    __angle_range = 2       #From -1 to 1    
    __radial_range = 0.5    #From 0 to 0.5
    __partitions = 0
    
    
    def __init__(self, feature_type, partitions, feature_info_path):
        '''
        Initialize spatial feature creation.
        @param feature_type: xy, radial or None
            xy: Creates quadtree like structure |  s = (x, y)
            radial: Circles and angles |  s = (r, sin(alpha), cos(alpha))
        @param partitions: Number of partitions (coarseness of quantization)
            if 0 values will be continuous
        @param feature_info_path: A file_path to store meta data - must be the same in all calls      
        '''
                
        if not (feature_type in self.__valid_features):
            raise ValueError("Argument descriptor_type is not valid. Must be in " + str(self.__valid_features))
        
        if not partitions >= 0:
            raise ValueError("Argument partitions must be integer and >= 0.")
        
        if feature_info_path == None:
            raise ValueError("Argument feature_info_path must not be none.")      
        
        # Set feature type
        # In case of SIFT there is no spatial feature, so set to None
        self.__feature_type = feature_type
        
        # Set quantization value
        if not isinstance(partitions, types.ListType):
            partitions = [partitions]
        self.__partitions = partitions
        
        self.__feature_info_path = feature_info_path
        return
    
    
    def __get_feature_sum(self, descriptors):
        '''Returns the average sum of an appearance feature.
           This info is stored as meta data in the dump path'''
        
        
        serializer = SerializationIO()
        
        if os.path.exists(self.__feature_info_path):
            meta_data = serializer.load_data(self.__feature_info_path)
            if meta_data.has_key("av_descriptor_sum"):
                return meta_data["av_descriptor_sum"]
        
        #Compute value and store data
        av_descriptor_sum = np.mean(np.sum(descriptors, axis=1))
        meta_data = dict()
        meta_data["av_descriptor_sum"] = av_descriptor_sum
        serializer.store_data(meta_data, self.__feature_info_path)
        return av_descriptor_sum
    
    def __get_polar_quantization(self, localizations, descriptors, partitions, reference_point=None):
        '''Compute a polar spatial feature that can be appended to the appearance feature descriptor.
           s = (r, sin(alpha), cos(alpha)) with 
           r being the radius to the reference_point and alpha the angle to an upward axis
           @param localizations: List of xy coordinates
           @param descriptors: List of feature descriptors
           @param partitions: Number of partitions 
           @param reference_point: Reference point for the distance computation, 
                                   e.g. center of image or patch.
           @return: Spatial feature.''' 
        
        #
        # Set length parameters
        #
        
        x_length = np.max(localizations[:,0])  - np.min(localizations[:,0])
        y_length = np.max(localizations[:,1])  - np.min(localizations[:,1])
        diag = ((x_length**2 + y_length**2)**(0.5))
                
        number_features = descriptors.shape[0]
        len_spatial = 3
        
        # Compute quantization step based on the number of partitions
        quantization_r = self.__radial_range / partitions
        quantization_sc = self.__angle_range / partitions
        
        # Compute quantization borders
        # => This will cause a radial setup with p tiles along the radial axis
        # => As well as p tiles along x and y (based on sin/cos)
        borders_sc = np.zeros(partitions-1)
        borders_r = np.zeros(partitions-1)
        for i in range(partitions-1):
            borders_sc[i] = -1 + (i+1)*quantization_sc
            borders_r[i] = 0 + (i+1)*quantization_r
        
        #
        # For each local feature compute
        # radius, sin(x-1), cos
        #
        
        #Vector to upward axis [0, 1]
        vec_up = np.arange(2)
        
        # Compute distances to reference points
        rmat = cdist(localizations, [reference_point])
        rmat = rmat.T[0]
        
        # Compute unit vectors to reference point
        vmat = localizations - reference_point
        # Avoid division by zero by array slicing
        div_idx = np.array(rmat) != 0
        vmat[div_idx] = vmat[div_idx] / np.array([rmat]).T[div_idx]
  
        # Compute radius to image size     
        rmat /= diag
        
        # Compute cosine and sin
        cosmat = np.dot(vmat,vec_up)
        
        # Note angles are only up to 180 degress and symmetric!
        # [0-Pi] only sees upper half of sin curve
        sinmat = np.sin(np.arccos(cosmat))
        # Offset sines to avoid mirroring along Y axis!
        sinmat[(vmat[:,0] < 0)] *= -1
        
        # Allocate space for the spatial feature vectors
        spatial_feature = np.zeros((number_features, len_spatial))
        
        # Increment radius according to border
        for br in borders_r:
            idx = (rmat >= br)
            spatial_feature[idx, 0] += 1
        
        # Increment sin, cos according to borders
        for bsc in borders_sc:
            idx = (sinmat >= bsc)
            spatial_feature[idx, 1] += 1
            idx = (cosmat >= bsc)
            spatial_feature[idx, 2] += 1
    
        #Return spatial feature
        return spatial_feature

    
    def __get_quadtree_quantization(self, localizations, descriptors, partitions, reference_point=None):
        '''Computes quantized xy coordinates that can be appended to the appearance feature descriptor.
           s = (q(x),q(y))
           @param localizations: List of xy coordinates
           @param descriptors: List of feature descriptors
           @param partitions: Number of partitions 
           @param reference_point: Reference point for the distance computation, 
                                   e.g. center of image or patch.
           @return: Spatial Feature.''' 
        
        
        #
        # Set length parameters
        #
        x_min = min(localizations[:,0])
        y_min = min(localizations[:,1])
        x_steps = (max(localizations[:,0])  - x_min) / partitions
        y_steps = (max(localizations[:,1])  - y_min) / partitions
        
        number_features = len(descriptors)
        len_spatial = 2
        
        #Create steps for xy-subregions
        borders = np.zeros((partitions-1, 2))
        for p in range(partitions)[1:]:
            borders[p-1] = np.array([x_min + p*x_steps, y_min + p*y_steps])
        
        #Allocate space for the spatial feature vectors
        spatial_feature = np.zeros((number_features, len_spatial))
        
        #Increment quantized xy-coordinates according to borders
        for b in borders:
            idx = (localizations[:,0] >= b[0])
            spatial_feature[idx, 0] += 1
            idx = (localizations[:,1] >= b[1])
            spatial_feature[idx, 1] += 1
        
        #Return spatial feature
        return spatial_feature
    
    def get_len_spatial_component(self):
        '''
        @return: Returns the length of the spatial component
        '''
        len_s = 2 # xy or lang_model encoding
        
        # len_s = 3 for radial part
        if self.__feature_type == "radial":
            len_s += 1
        
        # If multiple levels >> Additional coordinate
        if len(self.__partitions) > 1:
            len_s += 1
        return len_s 
    
    def add_spatial_feature(self, localizations, descriptors, reference_point=None, randomize=False):
        '''Appends spatial feature to the appearance feature descriptor:  v = (a_0, .. , a_n, s_0, .. , s_n)
           The spatial feature can be either quadtree quantized or radial quantized.
           @param localizations: List of xy coordinates
           @param descriptors: List of feature descriptors
           @param reference_point: Reference point for the distance computation, 
                                   e.g. center of image or patch.
           @param randomize: Shuffle the list of descriptors after modifying them.
                             In training this may be necessary to discard a spatial ordering.
           @return: Modified descriptor v.''' 
        
        
        # No spatial configurations required return old data
        if self.__feature_type == None:
            return descriptors
        
        # Set average feature sum, if not set by previous run
        if self.__feature_average_sum == None:
            self.__feature_average_sum = self.__get_feature_sum(descriptors)
        
        # Normalize descriptor to approx. sum one
        # Note: Do not use /= this does not work !?
        descriptors = descriptors / self.__feature_average_sum
                     
        # Compute artificial reference point if none given
        if reference_point == None:
            x_ref = (max(localizations[:,0])  - min(localizations[:,0]))/2 + min(localizations[:,0])
            y_ref = (max(localizations[:,1])  - min(localizations[:,1]))/2 + min(localizations[:,1])
            reference_point = np.array([x_ref, y_ref])
        else:
            # Make sure the reference point is a numpy array
            reference_point = np.array(reference_point)
        
        #Create target array
        n_d = len(descriptors)
        l_d = len(descriptors[0])
        l_s = 3
        if self.__feature_type == "radial":
            l_s = 4
        spatial_descriptors = np.zeros((len(self.__partitions)*n_d, l_d + l_s))
        
        i = 0
        for p in self.__partitions:
            # Compute spatial features
            if self.__feature_type == "radial":
                spatial_features = self.__get_polar_quantization(localizations, descriptors, p, reference_point)
            elif self.__feature_type == "xy":
                spatial_features = self.__get_quadtree_quantization(localizations, descriptors, p, reference_point)
            else:
                #If we get to this point something clearly went wrong!
                raise RuntimeWarning("No spatial features were computed - something might be wrong")
            #Create new feature vector of descriptor, spatial feature and pyramid level
            spatial_descriptors[i*n_d:(i+1)*n_d,:l_d] = descriptors
            spatial_descriptors[i*n_d:(i+1)*n_d,l_d:-1] = spatial_features
            spatial_descriptors[i*n_d:(i+1)*n_d,-1] = i
            i += 1
        
        #Delete old descriptors
        del descriptors
        
        # Randomize set of descriptors
        # For clustering the spatial order should be discarded
        if randomize:
            np.random.shuffle(spatial_descriptors)
        
        # Return modified descriptor
        return spatial_descriptors
