'''
Copyright (C) 2015 Rene Grzeszick. All rights reserved.
This file made available under the terms of the BSD license (see the Copyright.txt).

Created on Jun 30, 2014
'''

from file_handling.os_wrapper import OsUtils

class SVWConfig(object):
    '''
    Configuration file Classification experiments on the Caltech database
    '''
    
    def __check_asserts(self):
        assert(self.root_path != None)
        assert(self.classifier in ["SVM", "RandomForest", "Bayes"])
        assert(self.descriptor_type in ["SIFT", "SIFTxy", "SIFTrad"])
        assert(self.number_train_images > 0)
        assert(self.vocab_size > 0)        
        return
    
    def __init__(self,
                 root_path=None,
                 number_train_images=30,
                 descriptor_type="SIFT",
                 vocab_size=1000,
                 sift_step=5,
                 sift_sizes=(10),
                 spatial_partitions=(2),
                 rooted_histograms=True,
                 classifier="SVM"):
        '''
        Constructor initializes the parameters
        '''
        
        # Root path: 
        # The code expects a structure as follows: ROOT/data/[categories]
        #------------------------
        self.root_path = root_path
        
        # Training configuration
        #------------------------
        self.number_train_images = number_train_images
        
        
        # Bag-of-Features configuration
        #------------------------
        # Describes the Feature descriptor: SIFT|SIFTxy|SIFTradial
        self.descriptor_type = descriptor_type
        # Steps size of the sift grid
        self.sift_step = sift_step
        # Sizes for the descriptor
        self.sift_sizes = sift_sizes
        # Size of the visual vocabulary (centroids)
        self.vocab_size =  vocab_size
        #For spatial visual words: Partitions
        self.spatial_partitions = spatial_partitions
        # Square root the histogram (Hellinger Kernel)
        self.rooted_histograms = rooted_histograms
        # The python classifier
        self.classifier = classifier
        
        # Check assertions
        #------------------------
        self.__check_asserts()

        return
    
    
    # #################################
    # Descriptor specifications
    # #################################
    
    def get_descriptortype(self):
        '''
        Returns the type of the feature descriptor
        E.g. modified by spatial information
        '''
        return self.descriptor_type
    
    def get_spatial_feature_type(self):
        '''
        Returns the type of the spatial feature 
        that is encoded within the descriptor type
        '''
        spatial_feature_type = self.descriptor_type
        spatial_feature_type = spatial_feature_type.replace("SIFT", "")
        
        if spatial_feature_type == "":
            spatial_feature_type = None
        
        return spatial_feature_type
    

    
    # #################################
    # General paths
    # #################################
    
    def get_root_path(self):
        '''Returns the root path'''
        return self.root_path
   
    def data_path(self):
        '''Returns the data path where all images (train and test) are stored'''
        data_path = self.root_path + "/data"
        return data_path
    
    def train_path(self):
        '''Returns the train path'''
        return self.root_path + "/train"
          
    def test_path(self):
        '''Returns the test path'''
        return self.root_path + "/test"
    
    def feature_filename(self):
        '''Returns a filename for a feature file'''
        feature_type = self.__sift_parameterstring()
        return feature_type + ".feat"
    
    def feature_info_path(self):
        '''Returns a dump path for feature meta data'''
        feature_type = self.__sift_parameterstring()
        return self.data_path() + "/" + feature_type + "_feature_meta.info"
    
    def codebook_filename(self, feature_type=None):
        '''Returns the codelibrary filename for BoF representations'''
        feature_type = self.__sift_parameterstring()
        return  feature_type + ".cl"
    
    def classifier_filename(self):
        '''Returns the classifier model filename'''
        feature_type = self.__sift_parameterstring()
        return feature_type + "." + self.classifier.lower()
    
    # #################################
    # Pipeline parameterizations
    # Used for filepaths
    # ################################# 
    
    def __sift_parameterstring(self):
        '''Returns the sift data directory path'''
        grid_str = ""
        partition_str = ""
                
        #Descriptor string
        descr_str = self.descriptor_type
        
        #Spatial partitions
        if descr_str == "SIFTxy" or descr_str == "SIFTradial":
            partition_str += str(self.spatial_partitions) 
                
        #Grid sizes
        grid_str += "_" + str(self.sift_step) + "_"
        grid_str += str(self.sift_sizes)
               
        #return parameters            
        return descr_str + partition_str + grid_str  
    
    # #################################
    # Dataset helper methods
    # ################################# 
    
    def category_count(self):
        '''Counts all in the train Path categories'''
        trainPath = self.train_path()
        dir_list = OsUtils.get_dirlisting(trainPath, dirs_only=True)
        return len(dir_list)
    
    @staticmethod
    def __image_count(path):
        '''
        Counts images in a given path
        '''
        class_list = OsUtils.get_dirlisting(path, dirs_only=True)
        count = 0
        for class_name in class_list:
            file_list = OsUtils.get_dirlisting(path + "/" + class_name)
            count += len(file_list)
        return count
    
    
    def train_image_count(self):
        '''Returns the train image count'''
        train_path = self.root_path + "/train"     
        return self.__image_count(train_path)
    
    
    def all_image_count(self):
        '''Returns the data image count'''
        data_path = self.data_path()       
        return self.__image_count(data_path)
    
