'''
Copyright (C) 2015 Rene Grzeszick. All rights reserved.
This file made available under the terms of the BSD license (see the Copyright.txt).

Created on Jun 30, 2014
'''

from file_handling.os_wrapper import OsUtils
from file_handling.list_io import LineListIO
import random


class DatasetOrganizer(object):
    '''Organizes a dataset, e.g. performs a random split into test/train'''
    
    @staticmethod
    def split_rand_test_train(data_path, train_number, perc=None):
        '''
        @param data_path: Path that contains subfolderes (one for each class) storing the images
        @param train_number: Number of train images generated in the split
        @param perc: Use a fixed percentage of train images per class, 
                     the train_number is then used as an upper bound
        @return: List of training files
        '''
        
        # Get root path
        root_path = OsUtils.get_parent(data_path)
        
        # Create train and test dir
        OsUtils.recreate_dir(root_path + "/train")
        OsUtils.recreate_dir(root_path + "/test")
        train_path = root_path + "/train"
        test_path = root_path + "/test"
                
        # Iterate all classes
        class_list = OsUtils.get_dirlisting(data_path)
        trainfile_list = []
        for class_name in class_list:
            # Skip files
            if not OsUtils.is_dir(data_path, class_name):
                continue
            
            # Create train and test dir
            OsUtils.create_dir(train_path + "/" + class_name)
            OsUtils.create_dir(test_path + "/" + class_name)
            
            # Create train and test files
            file_list = OsUtils.get_dirlisting(data_path + "/" + class_name)
            random.shuffle(file_list)
            
            split_number = train_number
            if perc != None:
                if int(perc * len(file_list)) <  split_number:
                    split_number = int(perc * len(file_list))
            
            train_files = file_list[:split_number]
            test_files = file_list[split_number:]
            
            # Copy train files
            for train_file in train_files:
                trainfile_list.append(class_name + "/" + train_file)
                src = data_path + "/" + class_name + "/" + train_file
                dst = train_path + "/" + class_name + "/" + train_file
                OsUtils.copy_file(src, dst)                
            
            # Copy test files
            for test_file in test_files:
                src = data_path + "/" + class_name + "/" + test_file
                dst = test_path + "/" + class_name + "/" + test_file
                OsUtils.copy_file(src, dst)
                
        return trainfile_list
    
    @staticmethod
    def split_list_test_train(data_path, trainfile_list):
        '''
        @param data_path: Path that contains subfolderes (one for each class) storing the images
        @param train_number: Number of train images generated in the split
        @return: List of training files
        '''
        
        # Get root path
        root_path = OsUtils.get_parent(data_path)
        
        # Create train and test dir
        OsUtils.recreate_dir(root_path + "/train")
        OsUtils.recreate_dir(root_path + "/test")
        train_path = root_path + "/train"
        test_path = root_path + "/test"
                
        # Iterate all classes
        class_list = OsUtils.get_dirlisting(data_path)
        for class_name in class_list:
            # Skip files
            if not OsUtils.is_dir(data_path, class_name):
                continue
            
            # Create train and test dir
            OsUtils.create_dir(train_path + "/" + class_name)
            OsUtils.create_dir(test_path + "/" + class_name)
            
            # Create train and test files
            file_list = OsUtils.get_dirlisting(data_path + "/" + class_name)
            
            for file_name in file_list:
                if class_name + "/" + file_name in trainfile_list:
                    # Copy train file
                    src = data_path + "/" + class_name + "/" + file_name
                    dst = train_path + "/" + class_name + "/" + file_name
                    OsUtils.copy_file(src, dst)
                else:               
                    # Copy test files
                    src = data_path + "/" + class_name + "/" + file_name
                    dst = test_path + "/" + class_name + "/" + file_name
                    OsUtils.copy_file(src, dst)
                
        return
    
    @staticmethod
    def save_train_files(path, filename, trainfile_list):
        '''
        Stores the train file partitions that were used
        @param path: Path to store the traindata in
        @param filename: Filename (will be appended a unique number) 
        @param trainfile_list: List containing the relative paths to all training files  
        '''
        
        # Create unique filename
        i = 0
        while OsUtils.is_file(path, filename + str(i)):
            i += 1
        
        # Store trainfiles
        list_serializer = LineListIO()
        list_path = path + "/" + filename + "_" + str(i) 
        list_serializer.write_list(list_path, trainfile_list)
        
        return
    
    @staticmethod
    def load_train_files(path, filename):
        '''
        Loads the train file partitions that were used
        @param path: Path to store the traindata in
        @param filename: Filename (will be appended a unique number)
        '''
        list_serializer = LineListIO()
        list_path = path + "/" + filename
        trainfile_list = list_serializer.read_list(list_path)
        
        return trainfile_list
