'''
Copyright (C) 2015 Rene Grzeszick. All rights reserved.
This file made available under the terms of the BSD license (see the Copyright.txt).

Created on Jun 30, 2014
'''

import os, shutil, time, random

class OsUtils(object):
    '''
    This class contains various Utils for os operations.
    E.g. create / delete / list directories
    '''


    def __init__(self):
        '''
        Constructor
        '''
    @staticmethod
    def get_dirlisting(path, dirs_only=False, files_only=False):
        '''Returns a dirlisting that is alphabetically ordered'''
        dir_listing = os.listdir(path)
        dir_listing.sort()
        
        #Remove files from the dirlisting
        if dirs_only:
            dir_listing = [d for d in dir_listing
                           if os.path.isdir(path + "/" + d)]
            
        #Remove dirs from dirlisting
        if files_only:
            dir_listing = [d for d in dir_listing
                           if os.path.isfile(path + "/" + d)]

        return dir_listing
    
    @staticmethod
    def create_dir(path):
        '''Creates a directory if it does not already exist'''
        if not os.path.exists(path):
            os.mkdir(path)
        return
        
    @staticmethod
    def recreate_dir(path):
        '''Tries to remove a directory before creating it'''
        if os.path.exists(path):
            shutil.rmtree(path)
        time.sleep(0.1)
        OsUtils.create_dir(path)
        return
    
    @staticmethod  
    def delete_dir(path):
        '''Removes a directory tree with all sub dirs and files'''
        if os.path.exists(path):
            shutil.rmtree(path)
        return
    
    @staticmethod
    def delete_file(path):
        '''Removes a single file'''
        if os.path.exists(path):
            os.remove(path)
        return
    
    @staticmethod
    def is_dir(path, name=None):
        '''Returns whether the given name is a directory or not'''
        if name == None:
            return os.path.isdir(path)
        return os.path.isdir(path + "/" + name)
    
    @staticmethod
    def is_file(path, name=None):
        '''Returns whether the given name is a file or not'''
        if name == None:
            return os.path.isfile(path)
        return os.path.isfile(path + "/" + name)
    
    @staticmethod
    def get_parent(path):
        '''Returns the paths parent directory'''
        return os.path.split(path)[0]
    
    @staticmethod
    def get_dir_name(path):
        '''Returns the name of a file or directory without the complete path'''
        return os.path.split(path)[1]
    
    @staticmethod
    def get_random_file(path, filetype=None):
        '''Returns a random file from the given directory
        @param path:        directory path
        @param filetype:    optional filetype for the random file
        
        @return:            returns the complete filepath
        '''  
        if path == None:
            return None
        
        # Get dir listing
        dir_listing = os.listdir(path)
        random.shuffle(dir_listing)
        
        if filetype == None:
            # Search for any file
            for name in dir_listing:
                # Do not return directories
                if os.path.isfile(path + "/" + name):
                    return path + "/" + name
        else:
            # Search for a specific file type
            for name in dir_listing:
                if name.endswith(filetype):
                    return path + "/" + name
             
        return None
    
    @staticmethod
    def get_random_dir(path):
        '''Returns a random subfolder from the given directory
        @param path:        directory path
        
        @return:            returns the complete filepath
        '''
        if path == None:
            return None
        
        # Get dir listing
        dir_listing = os.listdir(path)
        random.shuffle(dir_listing)
        
        # Search for any directory
        for name in dir_listing:
            # Do not return files
            if os.path.isdir(path + "/" + name):
                return path + "/" + name
        
        return None
        
    @staticmethod
    def copy_file(src, dst):
        '''Copies a file from the src to the destination'''
        if os.path.exists(src):
            shutil.copyfile(src, dst)
        return
    
    @staticmethod
    def copy_dir(src, dst, ignore=None):
        '''Copies a complete dir and all subdirs/files in it'''
        if not os.path.exists(src) \
            or os.path.exists(dst):
            return
        
        if ignore == None:
            shutil.copytree(src, dst)
        else:
            def filelist_filter(filter_src,names):
                filter_names = [name for name in names
                                if '.' + name.split('.')[-1] in ignore]
                return filter_names
            shutil.copytree(src, dst, ignore=filelist_filter) 
        return
