'''
Created on Nov 12, 2012

@author   Mikael Rousson
@author:  rgrzeszi
'''

import _vlfeat
import numpy

def vl_sift(
        data, 
        frames=numpy.zeros(1), 
        octaves=-1, 
        levels=-1, 
        first_octave=0,
        peak_thresh=-1.0, 
        edge_thresh=-1.0,
        norm_thresh=-1.0,
        magnif=-1.0,
        window_size=-1.0,
        orientations=False,
        verbose=0):
    """ Computes the SIFT frames [1] (keypoints) F of the image I. I is a 
    gray-scale image in single precision. Each column of F is a feature frame 
    and has the format [X;Y;S;TH], where X,Y is the (fractional) center of the 
    frame, S is the scale and TH is the orientation (in radians). 
    Computes the SIFT descriptors [1] as well. Each column of D is the 
    descriptor of the corresponding frame in F. A descriptor is a 
    128-dimensional vector of class UINT8. 
    
    @param data         A gray-scale image in single precision 
                        (float numpy array).
    @param frames       Set the frames to use (bypass the detector). If frames 
                        are not passed in order of increasing scale, they are 
                        re-orderded. 
    @param octaves      Set the number of octave of the DoG scale space. 
    @param levels       Set the number of levels per octave of the DoG scale 
                        space. The default value is 3.
    @param first_octave Set the index of the first octave of the DoG scale 
                        space. The default value is 0.
    @param peak_thresh  Set the peak selection threshold. 
                        The default value is 0. 
    @param edge_thresh  Set the non-edge selection threshold. 
                        The default value is 10.
    @param norm_thresh  Set the minimum l2-norm of the descriptor before 
                        normalization. Descriptors below the threshold are set 
                        to zero.
    @param magnif       Set the descriptor magnification factor. The scale of 
                        the keypoint is multiplied by this factor to obtain the
                        width (in pixels) of the spatial bins. For instance, if
                        there are there are 4 spatial bins along each spatial
                        direction, the ``diameter'' of the descriptor is
                        approximatively 4 * MAGNIF. The default value is 3.
    @param orientations Compute the orientantions of the frames overriding the 
                        orientation specified by the 'Frames' option.
    @param verbose      Be verbose (may be repeated to increase the verbosity
                        level). 
    """
    if not data.flags['F_CONTIGUOUS']:
        data = numpy.array(data, order='F')        
    if not frames.flags['F_CONTIGUOUS']:
        frames = numpy.array(frames, order='F')
        
    return _vlfeat.vl_sift(data, frames, octaves, levels, first_octave, 
                        peak_thresh, edge_thresh, norm_thresh, magnif,
                        window_size, orientations, verbose)


def vl_dsift(
            data, 
            step=-1, 
            bounds=numpy.zeros(1, 'f'), 
            size=-1, 
            fast=True,
            windowsize=1.5, 
            verbose=False, 
            norm=False):
    """ [F,D] = VL_DSIFT(I) calculates the Dense Histogram of Gradients (DSIFT) 
    descriptors for the image I. I must be grayscale in SINGLE format.\n\n
    
    In this implementation, a DSIFT descriptor is equivalent to a SIFT 
    descriptor (see VL_SIFT()). This function calculates quickly a large number
    of such descriptors, for a dense covering of the image with features of the
    same size and orientation.\n\n
    
    The function returns the frames F and the descriptors D. Since all frames
    have identical size and orientation, F has only two rows (for the X and Y
    center coordinates). The orientation is fixed to zero. The scale is related
    to the SIZE of the spatial bins, which by default is equal to 3 pixels (see
    below). If NS is the number of bins in each spatial direction (by default
    4), then a DSIFT keypoint covers a square patch of NS by SIZE pixels.\n\n
    
    @remark The size of a SIFT bin is equal to the magnification factor MAGNIF 
    (usually 3) by the scale of the SIFT keypoint. This means that the scale of
    the SIFT keypoints corresponding to the DSIFT descriptors is SIZE / MAGNIF. 
    
    @remark Although related, DSIFT is not the same as the HOG descriptor used 
    in [1]. This descriptor is equivalent to SIFT instead. 
        
    @param step    Extract a descriptor each STEP pixels.
    @param size    A spatial bin covers SIZE pixels.
    @param norm    Append the frames with the normalization factor applied to 
                   each descriptor. In this case, F has 3 rows and this value 
                   is the 3rd row. This information can be used to suppress
                   descriptors with low contrast.
    @param fast    Use a flat rather than Gaussian window. Much faster.
    @param windowsize    Size of the Gaussian window
    @param verbose Be verbose. 
    """
    if not data.flags['F_CONTIGUOUS']:
        data = numpy.array(data, order='F')        
        
    return _vlfeat.vl_dsift(data, step, bounds, size, fast, windowsize, verbose, norm)


def vl_siftdescriptor(grad, frames):
    """ D = VL_SIFTDESCRIPTOR(GRAD, F) calculates the SIFT descriptors of the 
    keypoints F on the pre-processed image GRAD. GRAD is a 2xMxN array. The 
    first layer GRAD(1,:,:) contains the modulus of gradient of the original 
    image modulus. The second layer GRAD(2,:,:) contains the gradient angle 
    (measured in radians, clockwise, starting from the X axis -- this assumes 
    that the Y axis points down). The matrix F contains one column per keypoint 
    with the X, Y, SGIMA and ANLGE parameters. \n \n

    In order to match the standard SIFT descriptor, the gradient GRAD should be 
    calculated after mapping the image to the keypoint scale. This is obtained 
    by smoothing the image by a a Gaussian kernel of variance equal to the scale 
    of the keypoint. Additionaly, SIFT assumes that the input image is 
    pre-smoothed at scale 0.5 (this roughly compensates for the effect of the 
    CCD integrators), so the amount of smoothing that needs to be applied is 
    slightly less. The following code computes a standard SIFT descriptor by 
    using VL_SIFTDESCRIPTOR(): 
    """
    if not grad.flags['F_CONTIGUOUS']:
        grad = numpy.array(grad, order='F')
    if not frames.flags['F_CONTIGUOUS']:
        frames = numpy.array(frames, order='F')
        
    return _vlfeat.vl_siftdescriptor(grad, frames)

def vl_imsmooth(I, sigma):        
    """ I=VL_IMSMOOTH(I,SIGMA) convolves the image I by an isotropic Gaussian 
    kernel of standard deviation SIGMA. I must be an array of doubles. IF the 
    array is three dimensional, the third dimension is assumed to span different
    channels (e.g. R,G,B). In this case, each channel is convolved 
    independently.
    """
    if not I.flags['F_CONTIGUOUS']:
        I = numpy.array(I, order='F')
    return _vlfeat.vl_imsmooth(I, sigma)

       
def vl_rgb2gray(data):
    """ Rgb 2 gray consersion giving the same result as matlab own conversion 
    function.
    ATTENTION intensity values must be in [0..255]
    @param data A color image as 3D numpy array.
    @return A gray image as 2D numpy array (type is float but numbers are 
    rounded)    
    """
    if data.ndim == 3:
        return numpy.round(0.2989 * data[:,:,0] + 0.5870 * data[:,:,1] + 0.1140 * data[:,:,2])/255.0
    if data.ndim == 2:
        return data
    
    raise ValueError("Image cannot be converted 2 gray level. Unexpected dimensions: " + str(data.ndim))